\ DX-Forth library for access to dBase II DBF files.
\ Memory space or length of strings are NOT checked!
\ European date (DD/MM/YY) is used, edit this for US (MM/DD/YY)
\ ver. 3.1.1
\ Pavel Zampach, 2024
\ *

\ BUFLIB.F and DBFCONST.F must be included before!
marker DBFACC
inlib fileio

8 constant HEADSIZE
headsize buffer: HEADBUFF		\ dBase header buffer
integer DBFID				\ file ID of dBase file
variable CHANGED?			\ was file changed??

\ Number of records [n]
: #RECS		( -- n )
  [ headbuff 1+ ] literal @
;

\ Open DBF filename [addr u], extension isn't needed.
: OPEN-DBF	( addr u -- )
  s" DBF" +ext
  r/w open-file fc to dbfid
  headbuff headsize dbfid read-file fc drop
  $1a [ recbuff recsize + ] literal c!
  headbuff c@ 2 <> abort" no dBase II file"
  changed? off
;

\ Close DBF file.
: CLOSE-DBF	( -- )
  changed? @ if
[DEFINED] time&date [IF]
    time&date				\ European date DD/MM/YY used
    2000 - [ headbuff 5 + ] literal c!	\ year
    [ headbuff 4 + ] literal c!		\ month (3 + for MM/DD/YY)
    [ headbuff 3 + ] literal c!		\ day   (4 + for MM/DD/YY)
    drop 2drop				\ drop hour min sec
[THEN]
    0. dbfid reposition-file fc
    headbuff headsize dbfid write-file fc
  then
  dbfid close-file fc
;

\ True if record is "alive" (not marked as deleted).
: ALIVE?	( -- flag )
  recbuff c@ bl =
;

\ Set record's "alive" (not marked as deleted).
: SET-ALIVE	( flag -- )
  if bl else [char] * then
  recbuff c!
;

\ Set position to record [n] in DBF file.
: SET-POSITION	( n -- )
  #recs over > not abort" index too large"
  recsize m* 521. d+
  dbfid reposition-file fc
;

\ Internal component, write record [n] size [si].
: (WRITE-REC)	( si n -- )
  set-position
  recbuff swap dbfid write-file fc
  changed? on
;

\ Blank field [addr2 u2] and truncate string [addr1 u1] to fit
\ into this field.
: BLANK-TRIM	( addr1 u1 addr2 u2 -- addr1 u3 addr2 u2)
  2dup blank rot over min -rot
;

\ Read record [n] to record buffer,
\ records begin with zero!
: READ-REC	( n -- )
  set-position
  recbuff recsize dbfid read-file fc
  recsize <> abort" reading DBF error"
;

\ Write record [n] from record buffer,
\ records begin with zero!
: WRITE-REC	( n -- )
  recsize swap (write-rec)
;

\ Append new record from record buffer.
: APPEND-REC	( -- )
  true set-alive			\ set record "alive"
  recsize 1+ #recs			\ size with trailing EOF
  1 [ headbuff 1+ ] literal +!
  (write-rec)
;

\ Get [addr] of field [addr u].
aka drop GET-FIELDA	( addr u -- addr )

\ Dummy word for backward compatibility
aka noop GET-FIELDC	( addr u -- addr u )

\ Get string [addr2 u2] of field [addr1 u] with removed blanks
aka -blanks GET-FIELDCT	( addr1 u1 -- addr2 u2 )

\ Get double number [d] from field [addr u].
: GET-FIELDN	( addr u -- d )
  get-fieldct number? not if 0. then
;

\ Get logical value from field [addr u].
code GET-FIELDL	( addr u -- flag )
	H	pop
	H	pop
	M A	mov
	true H	lxi
	upc	call	\ A to upper case
	char Y	cpi
	hpush	jz
	char T	cpi
	hpush	jz
	char 1	cpi
	hpush	jz
	H	inx	\ false -> HL
		1push
end-code

\ Set string [addr1 u1] padding right to field [addr2 u2].
: SET-FIELDCR	( addr1 u1 addr2 u2 -- )
  blank-trim   + over - swap cmove
;

\ Set string [addr1 u1] padding left to field [addr2 u2].
: SET-FIELDCL	( addr1 u1 addr2 u2 -- )
  blank-trim   drop swap cmove
;

\ Set double number [d] to field [addr u].
: SET-FIELDN	( d addr u -- )
  2swap (d.) 2swap set-fieldcr
;

\ Set logical value [flag] to field [addr u].
: SET-FIELDL	( flag addr u -- )
  drop swap
  if [char] T else [char] F then
  swap c!
;

behead headsize changed?
behead set-position blank-trim
