\ Build DBFCONST.F from dBase II header for DBFACC v.3.1+
\ Using: DBFHEAD ufn[.dbf]
\ Pavel Zampach, 2023
\ *

inlib struct
inlib fileio
inlib applib
decimal

limit constant BUF
521 constant HEAD-LEN			\ length of dBase II header
$0d constant END-MARK			\ mark of end of header
buf 8 + value F-BASE			\ addr of field base
variable DBFOFF				\ offset from record begin
  1 dbfoff !				\ the first byte is deletion mark

<struct					\ Field descriptor:
  11 field ->F-NAME			\ ASCIIZ field name
   1 field ->D-TYPE			\ data type ('C'|'N'|'L')
   1 field ->F-LEN			\ field length
   2 field ->F-ADDR			\ field data address
   1 field ->DEC-LEN			\ number of decimal places
 struct> DESC-LEN			\ length of descriptor

\ Append comment
: APPCOM   ( addr u -- addr u+2 )
  [char] \ appc appbl
;

\ Add [x] to INTEGER or VALUE "name"
system
: +TO		( x "name" -- )
  postpone addr
  postpone +!
; immediate
application

\ Convert null terminated string [addr] to string [addr u]
: ZCOUNT   ( addr -- addr u )
  dup $ff 0 scan drop over -
;

\ Build complete filename
: FNAME   ( -- addr u )
  cmdtail s" DBF" +ext
;

\ Write string [addr u] to FID file
: WRITE-STR   ( addr u -- )
  fid write-line fc
;

\ Write begin of file
: WRITE-START   ( -- )
  <pad
  appcom fname apps write-str
  s" DECIMAL" write-str
  <pad
  [ buf 6 + ] literal @ appn appbl	\ record size
  s" CONSTANT RECSIZE" apps write-str
  s" RECSIZE 1+ BUFFER: RECBUFF" write-str
;

\ Write 2CONSTANT with address in RECBUFF and size
: WRITE-FIELD   ( -- )
  <pad
  dbfoff @ appn appbl			\ add offset
  s" RECBUFF + " apps
  f-base ->f-len c@
  dup dbfoff +!				\ update offset
  appn appbl				\ add length
  s" 2CONSTANT " apps
  f-base ->f-name zcount apps appbl	\ add name
  appcom f-base ->d-type c@ appc	\ add comment and data type
  f-base ->dec-len c@ ?dup if appn then	\ add decimal places if any
  write-str
;

\ Load DBF header
: LOAD-HEAD   ( -- )
  fname r/o open-file fc to fid
  buf head-len fid read-file fc drop
  fid close-file fc
  buf c@ 2 <> abort" no dBase II file"
;

: MAIN   ( -- )
  load-head
  s" DBFCONST.F" w/o create-file fc to fid
  write-start
  begin
    f-base c@ end-mark <>
  while
    write-field
    desc-len +to f-base			\ next field
  repeat
  fid close-file fc
;

turnkey main dbfhead
